/*
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 841 $ $LastChangedDate: 2002-09-01 17:22:49 +0200 (Sun, 01 Sep 2002) $
 *
 * Cg_global - code generation for global symbols. 
 */

open Ast;
open Util;
open Tyutil;

open Cg_util;
open Cg_types;

<Ksi::expr>list static_inits = [];

void push_union(var_symbol u)
{
    int cnt = 1;
    
    var_symbol get_name(<var_symbol>field f)
    {
	return get_binding(f.fld_symbol);
    }

    *(Ksi::id, int) get_enum_entry(var_symbol s)
    {
	set_property(s, Is_prototyped);
	return (make_id(s.a.extern_name), cnt++);
    }
    
    switch (union_of(u)).stype {
    case Union(lst):
	let fields = List::map(get_name, lst) in
	let content = List::map(get_enum_entry, fields) in
	output_decl(Ksi::Enum_def(content));
    case _:
	ice("Codegen::push_union(): not union");
    }
}

void push_proto(var_symbol s)
{
    if (has_property(s, Is_prototyped))
	return;

    if (has_property(s, Is_exception)) {
	let ex = List::cons1(Ksi::Extern) in
	output_decl(Ksi::Global(ex, ptroff, make_id(s.a.extern_name)));
    } else if (has_property(s, Is_union_field)) {
	push_union(s);
    } else {
	bool is_xc = has_property(s, Is_extern_c);
	bool is_public = is_xc || has_property(s, Is_exported) || 
			 s.a.fnc_level < 0;

	let attrs = public_attrs(is_public) in
	let name = make_id(s.a.extern_name) in
	if (is_global_function(s)) {
	    Ksi::ty proto_parm(*(ty, <id>option) p)
	    {
		let (t, _) = p in
		return transport_type(t);
	    }

	    _ ft = get_ft_record(type_instance(s.stype));
	    _ rt;

	    switch type_instance(ft.ret_type) {
	    case Void:
		rt = Ksi::Void;
	    case _:
		rt = transport_type(ft.ret_type);
	    }
	    
	    let parms = List::map(proto_parm, ft.parms) in 
	    let decl = Ksi::Func_proto(attrs, rt, name, parms) in
	    output_decl(decl);
	} else {
	    let t = storage_type(s.stype) in
	    if (!is_xc && s.module == Util::current_module)
		output_decl(Ksi::Global(attrs, t, name));
	    else
		let ex = Cons(Ksi::Extern, attrs) in
		output_decl(Ksi::Global(ex, t, name));
	}
    }

    set_property(s, Is_prototyped);
}

void push_static_lambda(string fnc_name, string lambda_name)
{
    let t = Ksi::Struct(make_id("gont_lambda_expr")) in
    let id = make_id(lambda_name) in
    let gd = Ksi::Global([Ksi::Local], t, id) in 
    let ptrref = Ksi::Component(Ksi::Ref(id), make_id("func_ptr")) in
    let ptr = Ksi::Convert(Ksi::Poly, 
			   Ksi::Func_ref(make_id(fnc_name))) in
    let ini = Ksi::Modify(ptrref, ptr) in {
	output_decl(gd);
	static_inits = Cons(ini, static_inits);
    }
}

Ksi::expr static_lambda(string name)
{
    string sname = genname(name + "$scl");
    push_static_lambda(name, sname);
    return Ksi::Unary(Ksi::Addr, Ksi::Ref(make_id(sname)));
}

Ksi::expr global_ref(var_symbol s)
{
    push_proto(s);

    Ksi::decl closurize()
    {
	int cnt = 0;
	
	*(Ksi::ty, Ksi::id) cl_parm(*(ty, <id>option) p)
	{
	    let (t, _) = p in
	    return (transport_type(t), make_id("_" + itoa(cnt++)));
	}

	int cnt2 = 0;
	
	Ksi::expr ref_parm(_)
	{
	    return Ksi::Ref(make_id("_" + itoa(cnt2++)));
	}

	_ ft = get_ft_record(s.stype);
	_ parms = Cons((null_type, dummy_id), List::map(cl_parm, ft.parms));
	_ rt = transport_type(ft.ret_type);
	_ name = make_id(s.a.extern_name + "$cl");
	
	_ body = Ksi::Call(make_id(s.a.extern_name), 
			   List::map(ref_parm, ft.parms));
				       
	switch type_instance(ft.ret_type) {
	case Void:
	    rt = Ksi::Void;
	case _:
	    body = Ksi::Return(body);
	}

	body = Ksi::Bind(body);

	return Ksi::Func(Nil, rt, name, parms, body);
    }

    if (has_property(s, Is_union_field))
	return ice("Codegen::global_ref(): union field requested");

    if (is_global_function(s)) {
	if (!has_property(s, Is_closurized)) {
	    output_decl(closurize());
	    set_property(s, Is_closurized);
	    let xn = s.a.extern_name in
	    push_static_lambda(xn + "$cl", xn + "$scl");
	}
	let ref = Ksi::Ref(make_id(s.a.extern_name + "$scl")) in
	return Ksi::Unary(Ksi::Addr, ref);
    } else {
	return Ksi::Ref(make_id(s.a.extern_name));
    }
}

<Ksi::expr>list get_global_inits()
{
    return static_inits;
}

Ksi::expr constant_union_expr(var_symbol s)
{
    push_proto(s);

    if (!has_property(s, Is_statically_allocated)) {
	let tag = symbol_union_tag(s) in
	let id = make_id(s.a.extern_name + "$sa") in
	let decl = Ksi::Global(List::cons1(Ksi::Local),
			       Ksi::Struct(make_id("gont_union")),
			       id) in
	let ini = Ksi::Modify(Ksi::Component(Ksi::Ref(id), make_id("tag")),
			      tag) in 
	{
	    output_decl(decl);
	    static_inits = Cons(ini, static_inits);
	    set_property(s, Is_statically_allocated);
	}
    }

    let id = make_id(s.a.extern_name + "$sa") in
    return Ksi::Unary(Ksi::Addr, Ksi::Ref(id));
}

// vim:sts=4:sw=4
